/*
   Implementation of a parallel algorithm that computes the product
   of an M x M sparse matrix and an M vector using one process for
   each of the N non-zero elements of the matrix. The implementation
   includes an assertion that checks that the final vector computed by
   the algorithm equals the vector obtained by multiplication according
   to the usual definition.

   To verify call Spin:
   > spin -a mm.pml && gcc -o mm pan.c && ./mm

   The implementation first sequentially initializes the matrix m and
   the vector x with nondeterministically chosen integer values in
   the range [-B..B]. Then it spawns N processes, which concurrently update
   one component of the output vector y each. Different processes may
   update the same component, but all the updates are specified to be atomic.
   When all updating processes terminate, the implementation checks that the
   value of vector y is what you would get by applying the standard matrix
   multiplication rule sequentially. If this check fails (for example, because
   you make the updates non-atomic), the model checker reports an assertion
   violation.

   The performance of the model checker verifying this implementation is largely
   insensitive to M and B since these values only affect the static initialization.
   In contrast, N determines the number of concurrent processes and hence the
   size of the global state space; for N > 10, the model checking run starts
   to consume significant amount of memory (and time).

   The analysis can be made more varied by trying out more complex  policies to
   assign updates to processes. For example, with a process pool and a work stealing
   queue.

   Author: Carlo A. Furia
*/

// number of non-zero elements of sparse matrix
#define N 10

// size of vectors
#define M 10
// if you change M's definition here, you have to copy the 
// new numeric value in the declaration of y's size below

// I can't find a way to use the #define M in C state declarations
c_state "int y[10]" "Global"
/* This C variable y represents the output vector. Since it
   need to be updated concurrently by different processes,
   it belongs to the global process state. */

// integer bounds for initial values: -B .. B
#define B 10

c_decl {
  typedef struct COO {
	 int row;
	 int col;
	 int val;
  } COO;
}

// This is read-only state, hence it need not be simulation state
c_code {
   COO m[N];
   /*
	 Every element e of m is a non-zero element of the matrix
	 at row e.row, column e.col, and value e.val.
   */
   int nassigned = 0;  // this is only used to initialize m
   int x[M];
}

c_code {
  // lookup element m(row, col) in sparse matrix m (with N non-zero elements)
  int lookup(int row, int col)
  {
    for (int i = 0; i < N; i++) {
	   COO el = m[i];
	   if (el.row == row && el.col == col)
		  return el.val;
	 }
	 // not found: it's zero
    return 0;
  }

  // return 1 iff check successful
  int check_multiply()
  {
    for (int i = 0; i < M; i++) {
	   int expect = 0;
	   for (int j = 0; j < M; j++)
		  expect += x[j] * lookup(j, i);
	   if (now.y[i] != expect)
		  return 0;
    }
	 return 1;
  }

  // return 1 iff m does NOT already have a component at row r and column c
  int available(int r, int c)
  {
    for (int i = 0; i < N && i < nassigned; i++)
      if (m[i].row == r && m[i].col == c)
         return 0;
    return 1;
  }
}

proctype setup()
{
  int k, r, c, v, av;
  // for every non-zero element
  for (k : 0 .. N-1) {
	 select (v : -B .. B);   // nondeterministically select value
	 // nondeterministically select row and column not already defined
reselect:
	 select (r : 0 .. M-1);
	 select (c : 0 .. M-1);
	 c_code{
		Psetup->av = available(Psetup->r, Psetup->c);
	 }
	 if
		:: av == 1 -> skip;
		:: else -> goto reselect;
	 fi
	 c_code {
		m[Psetup->k].row = Psetup->r;
		m[Psetup->k].col = Psetup->c;
		m[Psetup->k].val = Psetup->v;
		nassigned += 1;
	 };
  }
  // for every vector element
  for (k: 0 .. M-1) {
    select (v: -B .. B);  // nondeterministically select value
    c_code { x[Psetup->k] = Psetup->v; };
    // initialize result to zeros
	 c_code { now.y[Psetup->k] = 0; };
  }
}

// perform update of vector y corresponding to d-th component of m
proctype update(int d)
{
  int oldy;
  // removing the atomic block introduces races which lead to an assertion violation
  // note that each c_code is executed atomically; therefore, merging the two
  // blocks of c_code into one is equivalent to adding atomic
  atomic {
	 c_code {
		Pupdate->oldy = now.y[m[Pupdate->d].col];
	 }
	 c_code {
		now.y[m[Pupdate->d].col] = Pupdate->oldy  + (x[m[Pupdate->d].row] * m[Pupdate->d].val);
	 }
  }
}

proctype spawn_processes()
{
  int k;
  for (k: 0 .. N-1) {
	 run update(k);
  }
}

proctype check_result()
{
  int outcome;
  c_code {
	 Pcheck_result->outcome = check_multiply();
  };
  assert outcome == 1;
}

init {
  int w;
  w = _nr_pr;
  run setup();
  // block until setup has terminated
  (w == _nr_pr);
  w = _nr_pr;
  run spawn_processes();
  // block until all children have terminated
  (w == _nr_pr);
  run check_result();
}
